@@ -1,6 +1,6 @@
 # vim: set fileencoding=utf-8 :
 #
-# (C) 2006,2007,2008,2011 Guido Guenther <agx@sigxcpu.org>
+# (C) 2006,2007,2008,2011,2013 Guido Guenther <agx@sigxcpu.org>
 #    This program is free software; you can redistribute it and/or modify
 #    it under the terms of the GNU General Public License as published by
 #    the Free Software Foundation; either version 2 of the License, or
@@ -141,13 +141,15 @@ class GitRepository(object):
         cmd = ['git', command] + args
         env = cls.__build_env(extra_env)
         stderr_arg = subprocess.PIPE if capture_stderr else None
+        stdin_arg = subprocess.PIPE if input else None
 
         log.debug(cmd)
         popen = subprocess.Popen(cmd,
-                                 stdin=subprocess.PIPE,
+                                 stdin=stdin_arg,
                                  stdout=subprocess.PIPE,
                                  stderr=stderr_arg,
                                  env=env,
+                                 close_fds=True,
                                  cwd=cwd)
         (stdout, stderr) = popen.communicate(input)
         return stdout, stderr, popen.returncode
@@ -299,8 +301,10 @@ class GitRepository(object):
         """
         On what branch is the current working copy
 
-        @return: current branch
+        @return: current branch or C{None} in an empty repo
         @rtype: C{str}
+        @raises GitRepositoryError: if HEAD is not a symbolic ref
+          (e.g. when in detached HEAD state)
         """
         out, dummy, ret = self._git_inout('symbolic-ref', [ 'HEAD' ],
                                            capture_stderr=True)
@@ -308,12 +312,16 @@ class GitRepository(object):
             # We don't append stderr since
             # "fatal: ref HEAD is not a symbolic ref" confuses people
             raise GitRepositoryError("Currently not on a branch")
-
         ref = out.split('\n')[0]
+
         # Check if ref really exists
-        failed = self._git_getoutput('show-ref', [ ref ])[1]
-        if not failed:
-            return ref[11:] # strip /refs/heads
+        try:
+            self._git_command('show-ref', [ ref ])
+            branch = ref[11:] # strip /refs/heads
+        except GitRepositoryError:
+            branch = None  # empty repo
+        return branch
+
 
     def has_branch(self, branch, remote=False):
         """
@@ -329,8 +337,9 @@ class GitRepository(object):
             ref = 'refs/remotes/%s' % branch
         else:
             ref = 'refs/heads/%s' % branch
-        failed = self._git_getoutput('show-ref', [ ref ])[1]
-        if failed:
+        try:
+            self._git_command('show-ref', [ ref ])
+        except GitRepositoryError:
             return False
         return True
 
@@ -529,8 +538,18 @@ class GitRepository(object):
             if not self.has_branch(branch, remote=remote):
                 raise GitRepositoryError("Branch %s doesn't exist!" % branch)
 
-            self._git_inout('branch', ["--set-upstream", local_branch, upstream],
-                            capture_stderr=True)
+        if self._cmd_has_feature('branch', 'set-upstream-to'):
+            args = ['--set-upstream-to=%s' % upstream, local_branch]
+        else:
+            args = ["--set-upstream", local_branch, upstream]
+
+        dummy, err, ret = self._git_inout('branch',
+                                          args,
+                                          capture_stderr=True)
+        if ret:
+            raise GitRepositoryError(
+                "Failed to set upstream branch '%s' for '%s': %s" %
+                (upstream, local_branch, err.strip()))
 
     def get_upstream_branch(self, local_branch):
         """
